@@ -14,17 +14,19 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#include <config.h>
 #include <sys/time.h>
 #include <stdlib.h>
 #include <stdio.h>
-#include <config.h>
 
+#include "common/errcode.h"
 #include "common/evsched.h"
 
-/*! \todo Fix properly (issue #1581). */
-#ifndef HAVE_PSELECT
+/*! \bug #187: Disabled SLAB for events (testing reasons). */
+//#ifndef HAVE_PSELECT
 #define OPENBSD_SLAB_BROKEN
-#endif
+//#endif
+
 
 /*! \brief Some implementations of timercmp >= are broken, this is for compat.*/
 static inline int timercmp_ge(struct timeval *a, struct timeval *b) {
@@ -64,14 +66,15 @@ static void evsched_settimer(event_t *e, uint32_t dt)
 }
 
 /*! \brief Singleton application-wide event scheduler. */
-evsched_t *s_evsched = 0;
+evsched_t *s_evsched = NULL;
 
 evsched_t *evsched_new()
 {
 	evsched_t *s = malloc(sizeof(evsched_t));
 	if (!s) {
-		return 0;
+		return NULL;
 	}
+	memset(s, 0, sizeof(evsched_t));
 
 	/* Initialize event calendar. */
 	pthread_mutex_init(&s->rl, 0);
@@ -87,10 +90,7 @@ evsched_t *evsched_new()
 
 void evsched_delete(evsched_t **s)
 {
-	if (!s) {
-		return;
-	}
-	if (!*s) {
+	if (s == NULL || *s == NULL) {
 		return;
 	}
 
@@ -99,31 +99,32 @@ void evsched_delete(evsched_t **s)
 	pthread_mutex_destroy(&(*s)->mx);
 	pthread_cond_destroy(&(*s)->notify);
 
-	while (! EMPTY_HEAP(&(*s)->heap))	/* FIXME: Would be faster to simply walk through the array */
+#ifndef OPENBSD_SLAB_BROKEN
+	/* Free allocator (all events at once). */
+	slab_cache_destroy(&(*s)->cache.alloc);
+#else
+	while (! EMPTY_HEAP(&(*s)->heap))
 	{
 		event_t *e = *((event_t**)(HHEAD(&(*s)->heap)));
 		heap_delmin(&(*s)->heap);
 		evsched_event_free((*s), e);
 	}
-	
+#endif
+
 	free((*s)->heap.data);
 	(*s)->heap.data = NULL;;
 
-#ifndef OPENBSD_SLAB_BROKEN
-	/* Free allocator. */
-	slab_cache_destroy(&(*s)->cache.alloc);
-#endif
 	pthread_mutex_destroy(&(*s)->cache.lock);
 
 	/* Free scheduler. */
 	free(*s);
-	*s = 0;
+	*s = NULL;
 }
 
 event_t *evsched_event_new(evsched_t *s, int type)
 {
 	if (!s) {
-		return 0;
+		return NULL;
 	}
 
 	/* Allocate. */
@@ -141,6 +142,18 @@ event_t *evsched_event_new(evsched_t *s, int type)
 	/* Initialize. */
 	memset(e, 0, sizeof(event_t));
 	e->type = type;
+	e->parent = s;
+	return e;
+}
+
+event_t *evsched_event_new_cb(evsched_t *s, event_cb_t cb, void *data)
+{
+	/* Create event. */
+	event_t *e = evsched_event_new(s, EVSCHED_CB);
+	if (e != NULL) {
+		e->cb = cb;
+		e->data = data;
+	}
 	return e;
 }
 
@@ -163,7 +176,7 @@ event_t* evsched_next(evsched_t *s)
 {
 	/* Check. */
 	if (!s) {
-		return 0;
+		return NULL;
 	}
 
 	/* Lock calendar. */
@@ -183,7 +196,8 @@ event_t* evsched_next(evsched_t *s)
 
 			/* Immediately return. */
 			if (timercmp_ge(&dt, &next_ev->tv)) {
-				s->current = next_ev;
+				s->last_ev = next_ev;
+				s->running = true;
 				heap_delmin(&s->heap);
 				pthread_mutex_unlock(&s->mx);
 				pthread_mutex_lock(&s->rl);
@@ -204,67 +218,72 @@ event_t* evsched_next(evsched_t *s)
 
 	/* Unlock calendar, this shouldn't happen. */
 	pthread_mutex_unlock(&s->mx);
-	return 0;
-
+	return NULL;
 }
 
 int evsched_event_finished(evsched_t *s)
 {
 	if (!s) {
-		return -1;
+		return KNOT_EINVAL;
 	}
 
-	/* Mark as finished. */
-	if (s->current) {
-		s->current = 0;
-		pthread_mutex_unlock(&s->rl);
-		return 0;
+	/* \note This enables event cancellation & running on next event. */
+	if(s->running) {
+		if (pthread_mutex_unlock(&s->rl) != 0) {
+			return KNOT_ERROR;
+		}
+		s->running = false; /* Mark as not running. */
+	} else {
+		return KNOT_ENOTRUNNING;
 	}
 
-	/* Finished event is not current. */
-	return -1;
+	return KNOT_EOK;
 }
 
 int evsched_schedule(evsched_t *s, event_t *ev, uint32_t dt)
 {
 	if (!s || !ev) {
-		return -1;
+		return KNOT_EINVAL;
 	}
 
 	/* Update event timer. */
 	evsched_settimer(ev, dt);
 	ev->parent = s;
-	
+
 	/* Lock calendar. */
 	pthread_mutex_lock(&s->mx);
-	
+
+	/* Make sure it's not already enqueued. */
+	int found = 0;
+	if ((found = heap_find(&s->heap, ev))) {
+		heap_delete(&s->heap, found);
+	}
+
 	heap_insert(&s->heap, ev);
 
 	/* Unlock calendar. */
 	pthread_cond_broadcast(&s->notify);
 	pthread_mutex_unlock(&s->mx);
 
-	return 0;
+	return KNOT_EOK;
 }
 
 event_t* evsched_schedule_cb(evsched_t *s, event_cb_t cb, void *data, uint32_t dt)
 {
 	if (!s) {
-		return 0;
+		return NULL;
 	}
 
 	/* Create event. */
-	event_t *e = evsched_event_new(s, EVSCHED_CB);
+	event_t *e = evsched_event_new_cb(s, cb, data);
 	if (!e) {
-		return 0;
+		return NULL;
 	}
-	e->cb = cb;
-	e->data = data;
 
 	/* Schedule. */
 	if (evsched_schedule(s, e, dt) != 0) {
 		evsched_event_free(s, e);
-		e = 0;
+		e = NULL;
 	}
 
 	return e;
@@ -273,49 +292,80 @@ event_t* evsched_schedule_cb(evsched_t *s, event_cb_t cb, void *data, uint32_t d
 event_t* evsched_schedule_term(evsched_t *s, uint32_t dt)
 {
 	if (!s) {
-		return 0;
+		return NULL;
 	}
 
 	/* Create event. */
 	event_t *e = evsched_event_new(s, EVSCHED_TERM);
 	if (!e) {
-		return 0;
+		return NULL;
 	}
 
 	/* Schedule. */
 	if (evsched_schedule(s, e, dt) != 0) {
 		evsched_event_free(s, e);
-		e = 0;
+		e = NULL;
 	}
 
 	return e;
 }
 
-int evsched_cancel(evsched_t *s, event_t *ev)
+static int evsched_try_cancel(evsched_t *s, event_t *ev)
 {
-	int found;
+	int found = 0;
 
 	if (!s || !ev) {
-		return -1;
+		return KNOT_EINVAL;
 	}
-
-	/* Make sure not running. */
+	
+	/* Make sure not running. If an event is starting, we race for this lock
+	 * and either win or lose. If we lose, we may find it in heap because
+	 * it rescheduled itself. Either way, it will be marked as last running. */
 	pthread_mutex_lock(&s->rl);
-
+	
 	/* Lock calendar. */
 	pthread_mutex_lock(&s->mx);
-
+	
 	if ((found = heap_find(&s->heap, ev))) {
 		heap_delete(&s->heap, found);
 	}
 
+	/* Last running event was (probably) the one we're trying to cancel. */
+	if (s->last_ev == ev) {
+		s->last_ev = NULL;   /* Invalidate it. */
+		found = KNOT_EAGAIN; /* Let's try again if it didn't reschedule itself. */
+	}
+
 	/* Unlock calendar. */
 	pthread_cond_broadcast(&s->notify);
 	pthread_mutex_unlock(&s->mx);
-
+	
 	/* Enable running events. */
 	pthread_mutex_unlock(&s->rl);
 
-	return found;
+	if (found > 0) {        /* Event canceled. */
+		return KNOT_EOK;
+	} else if (found < 0) { /* Error */
+		return found;
+	}
+
+	return KNOT_ENOENT;     /* Not found. */
 }
 
+int evsched_cancel(evsched_t *s, event_t *ev)
+{
+	if (!s || !ev) {
+		return KNOT_EINVAL;
+	}
+
+	/* Event may have already run, try again. */
+	int ret = KNOT_EAGAIN;
+	while (ret == KNOT_EAGAIN) {
+		ret = evsched_try_cancel(s, ev);
+	}
+
+	/* Reset event timer. */
+	memset(&ev->tv, 0, sizeof(struct timeval));
+	/* Now we're sure event is canceled or finished. */
+	return KNOT_EOK;
+}
