@@ -43,8 +43,12 @@
 
 /* Constants. */
 #define ZONES_JITTER_PCT    10 /*!< +-N% jitter to timers. */
-#define IXFR_DBSYNC_TIMEOUT (60*1000) /*!< Database sync timeout = 60s. */
-#define AXFR_BOOTSTRAP_RETRY (60*1000) /*!< Interval between AXFR BS retries. */
+#define AXFR_BOOTSTRAP_RETRY (30*1000) /*!< Jitter cap between AXFR bootstrap retries. */
+#define AXFR_RETRY_MAXTIME (24*60*60*1000) /*!< Maximum AXFR retry interval cap of 24 hours. */
+
+enum {
+	REFRESH_DEFAULT = -1 /* Use time value from zone structure. */
+};
 
 /*!
  * \brief Zone-related data.
@@ -59,32 +63,29 @@ typedef struct zonedata_t
 
 	/*! \brief Zone data lock for exclusive access. */
 	pthread_mutex_t lock;
-	
+	/*! \brief Zone lock for DDNS. */
+	pthread_mutex_t ddns_lock;
+
 	/*! \brief Access control lists. */
 	acl_t *xfr_out;    /*!< ACL for xfr-out.*/
 	acl_t *notify_in;  /*!< ACL for notify-in.*/
 	acl_t *notify_out; /*!< ACL for notify-out.*/
+	acl_t *update_in; /*!< ACL for notify-out.*/
 
 	/*! \brief XFR-IN scheduler. */
 	struct {
-		acl_t         *acl;      /*!< ACL for xfr-in.*/
-		sockaddr_t     master;   /*!< Master server for xfr-in.*/
-		sockaddr_t     via;      /*!< Master server transit interface.*/
-		knot_key_t    tsig_key;  /*!< Master TSIG key. */
-		struct event_t *timer;   /*!< Timer for REFRESH/RETRY. */
-		struct event_t *expire;  /*!< Timer for REFRESH. */
-		pthread_mutex_t lock;    /*!< Pending XFR/IN lock. */
-		void           *wrkr;    /*!< Pending XFR/IN worker. */
-		int next_id;             /*!< ID of the next awaited SOA resp.*/
-		uint32_t bootstrap_retry;/*!< AXFR/IN bootstrap retry. */
-		unsigned       scheduled;/*!< Scheduled operations. */ 
+		acl_t          *acl;      /*!< ACL for xfr-in.*/
+		sockaddr_t      master;   /*!< Master server for xfr-in.*/
+		sockaddr_t      via;      /*!< Master server transit interface.*/
+		knot_tsig_key_t tsig_key; /*!< Master TSIG key. */
+		struct event_t *timer;    /*!< Timer for REFRESH/RETRY. */
+		struct event_t *expire;   /*!< Timer for REFRESH. */
+		uint32_t bootstrap_retry; /*!< AXFR/IN bootstrap retry. */
+		int has_master;           /*!< True if it has master set. */
+		unsigned state;
 	} xfr_in;
 
-	/*! \brief List of pending NOTIFY events. */
-	list notify_pending;
-	
-	/*! \brief List of fds with pending SOA queries. */
-	int soa_pending;
+	struct event_t *dnssec_timer;  /*!< Timer for DNSSEC events. */
 
 	/*! \brief Zone IXFR history. */
 	journal_t *ixfr_db;
@@ -92,31 +93,6 @@ typedef struct zonedata_t
 	uint32_t zonefile_serial;
 } zonedata_t;
 
-/*!
- * \brief Update zone database according to configuration.
- *
- * Creates a new database, copies references those zones from the old database
- * which are still in the configuration, loads any new zones required and
- * replaces the database inside the namserver.
- *
- * It also creates a list of deprecated zones that should be deleted once the
- * function finishes.
- *
- * This function uses RCU mechanism to guard the access to the config and
- * nameserver and to publish the new database in the nameserver.
- *
- * \param[in] conf Configuration.
- * \param[in] ns Nameserver which holds the zone database.
- * \param[out] db_old Old database, containing only zones which should be
- *                    deleted afterwards.
- *
- * \retval KNOT_EOK
- * \retval KNOT_EINVAL
- * \retval KNOT_ERROR
- */
-int zones_update_db_from_config(const conf_t *conf, knot_nameserver_t *ns,
-                               knot_zonedb_t **db_old);
-
 /*!
  * \brief Sync zone data back to text zonefile.
  *
@@ -138,9 +114,9 @@ int zones_zonefile_sync(knot_zone_t *zone, journal_t *journal);
 /*!
  * \todo Document me.
  */
-int zones_query_check_zone(const knot_zone_t *zone, const sockaddr_t *addr,
-                           knot_key_t **tsig_key, knot_rcode_t *rcode,
-                           uint8_t q_opcode);
+int zones_query_check_zone(const knot_zone_t *zone, uint8_t q_opcode,
+                           const sockaddr_t *addr, knot_tsig_key_t **tsig_key,
+                           knot_rcode_t *rcode);
 
 /*!
  * \todo Document me.
@@ -155,6 +131,14 @@ int zones_normal_query_answer(knot_nameserver_t *nameserver,
                               uint8_t *response_wire, size_t *rsize,
                               knot_ns_transport_t transport);
 
+/*!
+ * \todo Document me.
+ */
+int zones_process_update(knot_nameserver_t *nameserver,
+                         knot_packet_t *query, const sockaddr_t *addr,
+                         uint8_t *resp_wire, size_t *rsize,
+                         int fd, knot_ns_transport_t transport);
+
 /*!
  * \brief Processes normal response packet.
  *
@@ -169,7 +153,8 @@ int zones_normal_query_answer(knot_nameserver_t *nameserver,
  * \retval KNOT_EINVAL on invalid parameters or packet.
  * \retval KNOT_EMALF if an error occured and the response is not valid.
  */
-int zones_process_response(knot_nameserver_t *nameserver, 
+int zones_process_response(knot_nameserver_t *nameserver,
+                           int exp_msgid,
                            sockaddr_t *from,
                            knot_packet_t *packet, uint8_t *response_wire,
                            size_t *rsize);
@@ -202,7 +187,7 @@ int zones_ns_conf_hook(const struct conf_t *conf, void *data);
 /*!
  * \brief Store changesets in journal.
  *
- * Changesets will be stored on a permanent storage.
+ * Changesets will be stored to a permanent storage.
  * Journal may be compacted, resulting in flattening changeset history.
  *
  * \param zone Zone associated with the changeset.
@@ -215,7 +200,7 @@ int zones_ns_conf_hook(const struct conf_t *conf, void *data);
  * \todo Expects the xfr structure to be initialized in some way.
  * \todo Update documentation!!!
  */
-int zones_store_changesets(knot_ns_xfr_t *xfr);
+int zones_store_changesets(knot_zone_t *zone, knot_changesets_t *src, journal_t *j);
 
 /*!
  * \brief Begin changesets storing transaction.
@@ -223,7 +208,7 @@ int zones_store_changesets(knot_ns_xfr_t *xfr);
  * \retval pointer to journal if successful
  * \retval NULL on failure.
  */
-journal_t *zones_store_changesets_begin(knot_ns_xfr_t *xfr);
+journal_t *zones_store_changesets_begin(knot_zone_t *zone);
 
 /*!
  * \brief Commit stored changesets.
@@ -286,38 +271,123 @@ int zones_xfr_load_changesets(knot_ns_xfr_t *xfr, uint32_t serial_from,
  * \retval KNOT_ENODIFF when new zone's serial are equal.
  * \retval KNOT_ERROR when there was error creating changesets.
  */
-int zones_create_and_save_changesets(const knot_zone_t *old_zone,
-                                     const knot_zone_t *new_zone);
+int zones_create_changeset(const knot_zone_t *old_zone,
+                           const knot_zone_t *new_zone,
+                           knot_changeset_t *changeset);
+
+int zones_store_and_apply_chgsets(knot_changesets_t *chs,
+                                  knot_zone_t *zone,
+                                  knot_zone_contents_t **new_contents,
+                                  const char *msgpref, int type);
 
 /*!
  * \brief Update zone timers.
  *
  * REFRESH/RETRY/EXPIRE timers are updated according to SOA.
  *
- * \param sched Event scheduler.
  * \param zone Related zone.
- * \param cfzone Related zone contents. If NULL, configuration is
- *               reused.
+ * \param time Specific time or REFRESH_DEFAULT for default.
  *
  * \retval KNOT_EOK
  * \retval KNOT_EINVAL
  * \retval KNOT_ERROR
  */
-int zones_timers_update(knot_zone_t *zone, conf_zone_t *cfzone, evsched_t *sch);
+int zones_schedule_refresh(knot_zone_t *zone, int64_t time);
 
 /*!
- * \brief Cancel pending NOTIFY timer.
- *
- * \warning Expects locked zonedata lock.
- *
- * \param zd Zone data.
- * \param ev NOTIFY event.
+ * \brief Schedule NOTIFY after zone update.
+ * \param zone Related zone.
  *
  * \retval KNOT_EOK
  * \retval KNOT_ERROR
- * \retval KNOT_EINVAL
  */
-int zones_cancel_notify(zonedata_t *zd, notify_ev_t *ev);
+int zones_schedule_notify(knot_zone_t *zone);
+
+/*!
+ * \brief Cancel DNSSEC event.
+ *
+ * \param zone  Related zone.
+ *
+ * \return Error code, KNOT_OK if successful.
+ */
+int zones_cancel_dnssec(knot_zone_t *zone);
+
+/*!
+ * \brief Schedule DNSSEC event.
+ * \param zone Related zone.
+ * \param unixtime When to schedule.
+ * \param force Force sign or not
+ *
+ * \return Error code, KNOT_OK if successful.
+ */
+int zones_schedule_dnssec(knot_zone_t *zone, time_t unixtime);
+
+/*!
+ * \brief Schedule IXFR sync for given zone.
+ *
+ * \param zone            Zone to scheduler IXFR sync for.
+ * \param dbsync_timeout  Sync time in seconds.
+ */
+void zones_schedule_ixfr_sync(knot_zone_t *zone, int dbsync_timeout);
+
+/*!
+ * \brief Processes forwarded UPDATE response packet.
+ * \todo #1291 move to appropriate section (DDNS).
+ */
+int zones_process_update_response(knot_ns_xfr_t *data, uint8_t *rwire, size_t *rsize);
+
+/*!
+ * \brief Verify TSIG in query.
+ *
+ * \param query Query packet.
+ * \param key TSIG key used for this query.
+ * \param rcode Dst for resulting RCODE.
+ * \param tsig_rcode Dst for resulting TSIG RCODE.
+ * \param tsig_prev_time_signed Dst for previout time signed.
+ *
+ * \return KNOT_EOK if verified or error if not.
+ */
+int zones_verify_tsig_query(const knot_packet_t *query,
+                            const knot_tsig_key_t *key,
+                            knot_rcode_t *rcode, uint16_t *tsig_rcode,
+                            uint64_t *tsig_prev_time_signed);
+
+/*!
+ * \brief Apply changesets to zone from journal.
+ *
+ * \param zone Specified zone.
+ *
+ * \retval KNOT_EOK if successful.
+ * \retval KNOT_EINVAL on invalid parameters.
+ * \retval KNOT_ENOENT if zone has no contents.
+ * \retval KNOT_ERROR on unspecified error.
+ */
+int zones_journal_apply(knot_zone_t *zone);
+
+/*!
+ * \brief Creates diff and DNSSEC changesets and stores them to journal.
+ *
+ * \param z             Zone configuration.
+ * \param zone          Zone to sign.
+ * \param ns            Name server structure.
+ * \param zone_changed  The zone was loaded or modified.
+ *
+ * \return Error code, KNOT_OK if successful.
+ */
+int zones_do_diff_and_sign(const conf_zone_t *z, knot_zone_t *zone,
+                           const knot_nameserver_t *ns, bool zone_changed);
+
+/*! \brief Just sign current zone. */
+int zones_dnssec_sign(knot_zone_t *zone, bool force, uint32_t *expires_at);
+
+/*
+ * Event callbacks.
+ */
+
+int zones_expire_ev(event_t *e);
+int zones_refresh_ev(event_t *e);
+int zones_flush_ev(event_t *e);
+int zones_dnssec_ev(event_t *event);
 
 #endif // _KNOTD_ZONES_H_
 
