@@ -21,138 +21,57 @@
 
 #include <urcu.h>
 
-#include "common.h"
-#include "zone/node.h"
-#include "rrset.h"
-#include "common/skip-list.h"
-#include "common/tree.h"
-#include "util/debug.h"
+#include "libknot/common.h"
+#include "libknot/zone/node.h"
+#include "libknot/rrset.h"
+#include "common/descriptor.h"
+#include "libknot/util/debug.h"
 
 /*----------------------------------------------------------------------------*/
 /* Non-API functions                                                          */
 /*----------------------------------------------------------------------------*/
 /*!
- * \brief Returns the delegation point flag
+ * \brief Sets the given flag to node's flags.
  *
- * \param flags Flags to retrieve the flag from.
- *
- * \return A byte with only the delegation point flag set if it was set in
- *         \a flags.
+ * \param node Node to set the flag in.
+ * \param flag Flag to set.
  */
-static inline uint8_t knot_node_flags_get_deleg(uint8_t flags)
+static inline void knot_node_flags_set(knot_node_t *node, uint8_t flag)
 {
-	return flags & KNOT_NODE_FLAGS_DELEG;
+	node->flags |= flag;
 }
 
 /*----------------------------------------------------------------------------*/
 /*!
- * \brief Sets the delegation point flag.
+ * \brief Returns the given flag from node's flags.
  *
- * \param flags Flags to set the flag in.
+ * \param node Node to set the flag in.
+ * \param flag Flag to retrieve.
+ *
+ * \return A byte with only the given flag set if it was set in \a node.
  */
-static inline void knot_node_flags_set_deleg(uint8_t *flags)
+static inline uint8_t knot_node_flags_get(const knot_node_t *node, uint8_t flag)
 {
-	*flags |= KNOT_NODE_FLAGS_DELEG;
+	return node->flags & flag;
 }
 
 /*----------------------------------------------------------------------------*/
 /*!
- * \brief Clears the delegation point flag.
+ * \brief Clears the given flag in node's flags.
  *
- * \param flags Flags to clear the flag in.
+ * \param node Node to clear the flag in.
+ * \param flag Flag to clear.
  */
-static inline void knot_node_flags_clear_deleg(uint8_t *flags)
+static inline void knot_node_flags_clear(knot_node_t *node, uint8_t flag)
 {
-	*flags &= ~KNOT_NODE_FLAGS_DELEG;
-}
-
-/*----------------------------------------------------------------------------*/
-/*!
- * \brief Returns the non-authoritative node flag
- *
- * \param flags Flags to retrieve the flag from.
- *
- * \return A byte with only the non-authoritative node flag set if it was set in
- *         \a flags.
- */
-static inline uint8_t knot_node_flags_get_nonauth(uint8_t flags)
-{
-	return flags & KNOT_NODE_FLAGS_NONAUTH;
-}
-
-/*----------------------------------------------------------------------------*/
-/*!
- * \brief Sets the non-authoritative node flag.
- *
- * \param flags Flags to set the flag in.
- */
-static inline void knot_node_flags_set_nonauth(uint8_t *flags)
-{
-	*flags |= KNOT_NODE_FLAGS_NONAUTH;
-}
-
-/*----------------------------------------------------------------------------*/
-/*!
- * \brief Clears the non-authoritative node flag.
- *
- * \param flags Flags to clear the flag in.
- */
-static inline void knot_node_flags_clear_nonauth(uint8_t *flags)
-{
-	*flags &= ~KNOT_NODE_FLAGS_NONAUTH;
-}
-
-/*----------------------------------------------------------------------------*/
-/*!
- * \brief Sets the empty node flag.
- *
- * \param flags Flags to set the flag in.
- */
-static inline void knot_node_flags_set_empty(uint8_t *flags)
-{
-	*flags |= KNOT_NODE_FLAGS_EMPTY;
-}
-
-/*----------------------------------------------------------------------------*/
-/*!
- * \brief Returns the empty node flag
- *
- * \param flags Flags to retrieve the flag from.
- *
- * \return A byte with only the empty node flag set if it was set in \a flags.
- */
-static inline uint8_t knot_node_flags_get_empty(uint8_t flags)
-{
-	return flags & KNOT_NODE_FLAGS_EMPTY;
-}
-
-/*----------------------------------------------------------------------------*/
-/*!
- * \brief Compares the two keys as RR types.
- *
- * \note This function may be used in data structures requiring generic
- *       comparation function.
- *
- * \param key1 First RR type.
- * \param key2 Second RR type.
- *
- * \retval 0 if \a key1 is equal to \a key2.
- * \retval < 0 if \a key1 is lower than \a key2.
- * \retval > 0 if \a key1 is higher than \a key2.
- */
-static int compare_rrset_types(void *rr1, void *rr2)
-{
-	knot_rrset_t *rrset1 = (knot_rrset_t *)rr1;
-	knot_rrset_t *rrset2 = (knot_rrset_t *)rr2;
-	return ((rrset1->type > rrset2->type) ? 1 :
-	        (rrset1->type == rrset2->type) ? 0 : -1);
+	node->flags &= ~flag;
 }
 
 /*----------------------------------------------------------------------------*/
 /* API functions                                                              */
 /*----------------------------------------------------------------------------*/
 
-knot_node_t *knot_node_new(knot_dname_t *owner, knot_node_t *parent,
+knot_node_t *knot_node_new(const knot_dname_t *owner, knot_node_t *parent,
                            uint8_t flags)
 {
 	knot_node_t *ret = (knot_node_t *)calloc(1, sizeof(knot_node_t));
@@ -161,62 +80,81 @@ knot_node_t *knot_node_new(knot_dname_t *owner, knot_node_t *parent,
 		return NULL;
 	}
 
-	/* Store reference to owner. */
-	knot_dname_retain(owner);
-	ret->owner = owner;
+	/*! \todo This is inconsistent: knot_rrset_new() does not copy owner.
+	 *        Either copy in all _new() functions, or in none. I vote for
+	 *        the former, as it should be responsibility of the caller to
+	 *        do the copying (or not if he decides to do so).
+	 */
+	if (owner) {
+		ret->owner = knot_dname_copy(owner);
+	}
+
 	knot_node_set_parent(ret, parent);
-	ret->rrset_tree = gen_tree_new(compare_rrset_types);
+	ret->rrset_tree = NULL;
 	ret->flags = flags;
-	
+
 	assert(ret->children == 0);
 
 	return ret;
 }
 
-/*----------------------------------------------------------------------------*/
-/*! \todo Consider replacing rrset_merge() with rrset_merge_no_dupl(). Currently
- *        this function is never called with merge=1, so it's not a problem,
- *        but it may be in the future.
- */
-int knot_node_add_rrset(knot_node_t *node, knot_rrset_t *rrset,
-                        int merge)
+
+int knot_node_add_rrset_no_merge(knot_node_t *node, knot_rrset_t *rrset)
 {
 	if (node == NULL) {
 		return KNOT_EINVAL;
 	}
 
-	int ret = 0;
-
-	if ((ret = (gen_tree_add(node->rrset_tree, rrset,
-	                         (merge) ? knot_rrset_merge : NULL))) < 0) {
-		dbg_node("Failed to add rrset to node->rrset_tree.\n");
-		return KNOT_ERROR;
-	}
-
-	if (ret >= 0) {
-		node->rrset_count += (ret > 0 ? 0 : 1);
-		return ret;
-	} else {
-		return KNOT_ERROR;
+	size_t nlen = (node->rrset_count + 1) * sizeof(knot_rrset_t*);
+	void *p = realloc(node->rrset_tree, nlen);
+	if (p == NULL) {
+		return KNOT_ENOMEM;
 	}
+	node->rrset_tree = p;
+	node->rrset_tree[node->rrset_count] = rrset;
+	++node->rrset_count;
+	return KNOT_EOK;
 }
 
-int knot_node_add_rrset_no_dupl(knot_node_t *node, knot_rrset_t *rrset)
+int knot_node_add_rrset_replace(knot_node_t *node, knot_rrset_t *rrset)
 {
-	int ret = 0;
-
-	if ((ret = (gen_tree_add(node->rrset_tree, rrset,
-	                         knot_rrset_merge_no_dupl))) < 0) {
-		dbg_node("Failed to add rrset to node->rrset_tree.\n");
-		return KNOT_ERROR;
+	if (node == NULL) {
+		return KNOT_EINVAL;
 	}
 
-	if (ret >= 0) {
-		node->rrset_count += (ret > 0 ? 0 : 1);
-		return ret;
-	} else {
-		return KNOT_ERROR;
+	for (uint16_t i = 0; i < node->rrset_count; ++i) {
+		if (node->rrset_tree[i]->type == rrset->type) {
+		node->rrset_tree[i] = rrset;
+		}
 	}
+
+	return knot_node_add_rrset_no_merge(node, rrset);
+}
+
+int knot_node_add_rrset(knot_node_t *node, knot_rrset_t *rrset)
+{
+	if (node == NULL) {
+		return KNOT_EINVAL;
+	}
+
+	for (uint16_t i = 0; i < node->rrset_count; ++i) {
+		if (node->rrset_tree[i]->type == rrset->type) {
+			int merged, deleted_rrs;
+			int ret = knot_rrset_merge_sort(node->rrset_tree[i],
+			                                rrset, &merged,
+			                                &deleted_rrs);
+			if (ret != KNOT_EOK) {
+				return ret;
+			} else if (merged || deleted_rrs) {
+				return 1;
+			} else {
+				return 0;
+			}
+		}
+	}
+
+	// New RRSet (with one RR)
+	return knot_node_add_rrset_no_merge(node, rrset);
 }
 
 /*----------------------------------------------------------------------------*/
@@ -224,14 +162,7 @@ int knot_node_add_rrset_no_dupl(knot_node_t *node, knot_rrset_t *rrset)
 const knot_rrset_t *knot_node_rrset(const knot_node_t *node,
                                         uint16_t type)
 {
-	if (node == NULL) {
-		return NULL;
-	}
-
-	assert(node->rrset_tree != NULL);
-	knot_rrset_t rrset;
-	rrset.type = type;
-	return (const knot_rrset_t *)gen_tree_find(node->rrset_tree, &rrset);
+	return knot_node_get_rrset(node, type);
 }
 
 /*----------------------------------------------------------------------------*/
@@ -242,9 +173,14 @@ knot_rrset_t *knot_node_get_rrset(const knot_node_t *node, uint16_t type)
 		return NULL;
 	}
 
-	knot_rrset_t rrset;
-	rrset.type = type;
-	return (knot_rrset_t *)gen_tree_find(node->rrset_tree, &rrset);
+	knot_rrset_t **rrs = node->rrset_tree;
+	for (uint16_t i = 0; i < node->rrset_count; ++i) {
+		if (rrs[i]->type == type) {
+			return rrs[i];
+		}
+	}
+
+	return NULL;
 }
 
 /*----------------------------------------------------------------------------*/
@@ -255,15 +191,18 @@ knot_rrset_t *knot_node_remove_rrset(knot_node_t *node, uint16_t type)
 		return NULL;
 	}
 
-	knot_rrset_t dummy_rrset;
-	dummy_rrset.type = type;
-	knot_rrset_t *rrset =
-		(knot_rrset_t *)gen_tree_find(node->rrset_tree, &dummy_rrset);
-	if (rrset != NULL) {
-		gen_tree_remove(node->rrset_tree, rrset);
-		node->rrset_count--;
+	uint16_t i = 0;
+	knot_rrset_t *ret = NULL;
+	knot_rrset_t **rrs = node->rrset_tree;
+	for (; i < node->rrset_count && ret == NULL; ++i) {
+		if (rrs[i]->type == type) {
+			ret = rrs[i];
+			memmove(rrs + i, rrs + i + 1, (node->rrset_count - i - 1) * sizeof(knot_rrset_t *));
+			--node->rrset_count;
+		}
 	}
-	return rrset;
+
+	return ret;
 }
 
 /*----------------------------------------------------------------------------*/
@@ -275,9 +214,7 @@ void knot_node_remove_all_rrsets(knot_node_t *node)
 	}
 
 	// remove RRSets but do not delete them
-	gen_tree_clear(node->rrset_tree);
 	node->rrset_count = 0;
-
 }
 
 /*----------------------------------------------------------------------------*/
@@ -293,48 +230,19 @@ short knot_node_rrset_count(const knot_node_t *node)
 
 /*----------------------------------------------------------------------------*/
 
-struct knot_node_save_rrset_arg {
-	knot_rrset_t **array;
-	size_t count;
-	size_t max_count;
-};
-
-/*----------------------------------------------------------------------------*/
-
-static void save_rrset_to_array(void *node, void *data)
-{
-	struct knot_node_save_rrset_arg *args =
-		(struct knot_node_save_rrset_arg *)data;
-	knot_rrset_t *rrset = (knot_rrset_t *)node;
-
-	if (args->count > args->max_count) {
-		++args->count;
-		return;
-	}
-
-	args->array[args->count++] = rrset;
-}
-
-/*----------------------------------------------------------------------------*/
-
 knot_rrset_t **knot_node_get_rrsets(const knot_node_t *node)
 {
 	if (node == NULL || node->rrset_count == 0) {
 		return NULL;
 	}
-	knot_rrset_t **rrsets = (knot_rrset_t **)malloc(
-		node->rrset_count * sizeof(knot_rrset_t *));
-	CHECK_ALLOC_LOG(rrsets, NULL);
-	struct knot_node_save_rrset_arg args;
-	args.array = rrsets;
-	args.count = 0;
-	args.max_count = node->rrset_count;
 
-	gen_tree_apply_inorder(node->rrset_tree, save_rrset_to_array, &args);
+	size_t rrlen = node->rrset_count * sizeof(knot_rrset_t*);
+	knot_rrset_t **cpy = malloc(rrlen);
+	if (cpy != NULL) {
+		memcpy(cpy, node->rrset_tree, rrlen);
+	}
 
-	assert(args.count == node->rrset_count);
-
-	return rrsets;
+	return cpy;
 }
 
 /*----------------------------------------------------------------------------*/
@@ -348,31 +256,21 @@ const knot_rrset_t **knot_node_rrsets(const knot_node_t *node)
 	return (const knot_rrset_t **)knot_node_get_rrsets(node);
 }
 
-/*----------------------------------------------------------------------------*/
-
-static void count_rrsets(void *node, void *data)
-{
-	assert(node != NULL);
-	assert(data != NULL);
-
-	int *count = (int *)data;
-	++(*count);
-}
-
-/*----------------------------------------------------------------------------*/
-
-int knot_node_count_rrsets(const knot_node_t *node)
+knot_rrset_t **knot_node_get_rrsets_no_copy(const knot_node_t *node)
 {
 	if (node == NULL) {
-		return KNOT_EINVAL;
+		return NULL;
 	}
 
-	int count = 0;
-	gen_tree_apply_inorder(node->rrset_tree, count_rrsets, (void *)&count);
-
-	return count;
+	return node->rrset_tree;
 }
 
+const knot_rrset_t **knot_node_rrsets_no_copy(const knot_node_t *node)
+{
+	return (const knot_rrset_t **)knot_node_get_rrsets_no_copy(node);
+}
+
+
 /*----------------------------------------------------------------------------*/
 
 const knot_node_t *knot_node_parent(const knot_node_t *node)
@@ -409,7 +307,7 @@ void knot_node_set_parent(knot_node_t *node, knot_node_t *parent)
 	}
 	// set the parent
 	node->parent = parent;
-	
+
 	// increase the count of children of the new parent
 	if (parent != NULL) {
 		++parent->children;
@@ -445,7 +343,7 @@ knot_node_t *knot_node_get_previous(const knot_node_t *node)
 	if (node == NULL) {
 		return NULL;
 	}
-	
+
 	return node->prev;
 }
 
@@ -467,7 +365,7 @@ knot_node_t *knot_node_get_nsec3_node(const knot_node_t *node)
 	if (node == NULL) {
 		return NULL;
 	}
-	
+
 	return node->nsec3_node;
 }
 
@@ -491,9 +389,6 @@ void knot_node_set_nsec3_node(knot_node_t *node, knot_node_t *nsec3_node)
 	}
 
 	node->nsec3_node = nsec3_node;
-	if (nsec3_node != NULL) {
-		nsec3_node->nsec3_referer = node;
-	}
 }
 
 /*----------------------------------------------------------------------------*/
@@ -516,18 +411,6 @@ knot_dname_t *knot_node_get_owner(const knot_node_t *node)
 
 /*----------------------------------------------------------------------------*/
 
-void knot_node_set_owner(knot_node_t *node, knot_dname_t* owner)
-{
-	if (node) {
-		/* Retain new owner and release old owner. */
-		knot_dname_retain(owner);
-		knot_dname_release(node->owner);
-		node->owner = owner;
-	}
-}
-
-/*----------------------------------------------------------------------------*/
-
 knot_node_t *knot_node_get_wildcard_child(const knot_node_t *node)
 {
 	if (node == NULL) {
@@ -597,28 +480,6 @@ void knot_node_set_new_node(knot_node_t *node,
 
 /*----------------------------------------------------------------------------*/
 
-void knot_node_set_zone(knot_node_t *node, const knot_zone_t *zone)
-{
-	if (node == NULL) {
-		return;
-	}
-
-	node->zone = zone;
-}
-
-/*----------------------------------------------------------------------------*/
-
-const knot_zone_t *knot_node_zone(const knot_node_t *node)
-{
-	if (node == NULL) {
-		return NULL;
-	}
-
-	return node->zone;
-}
-
-/*----------------------------------------------------------------------------*/
-
 void knot_node_update_ref(knot_node_t **ref)
 {
 	if (*ref != NULL && (*ref)->new_node != NULL) {
@@ -638,8 +499,6 @@ void knot_node_update_refs(knot_node_t *node)
 	knot_node_update_ref(&node->wildcard_child);
 	// reference to NSEC3 node
 	knot_node_update_ref(&node->nsec3_node);
-	// reference to NSEC3 referrer
-	knot_node_update_ref(&node->nsec3_referer);
 }
 
 /*----------------------------------------------------------------------------*/
@@ -650,7 +509,7 @@ void knot_node_set_deleg_point(knot_node_t *node)
 		return;
 	}
 
-	knot_node_flags_set_deleg(&node->flags);
+	knot_node_flags_set(node, KNOT_NODE_FLAGS_DELEG);
 }
 
 /*----------------------------------------------------------------------------*/
@@ -661,7 +520,7 @@ int knot_node_is_deleg_point(const knot_node_t *node)
 		return KNOT_EINVAL;
 	}
 
-	return knot_node_flags_get_deleg(node->flags);
+	return knot_node_flags_get(node, KNOT_NODE_FLAGS_DELEG);
 }
 
 /*----------------------------------------------------------------------------*/
@@ -672,7 +531,7 @@ void knot_node_set_non_auth(knot_node_t *node)
 		return;
 	}
 
-	knot_node_flags_set_nonauth(&node->flags);
+	knot_node_flags_set(node, KNOT_NODE_FLAGS_NONAUTH);
 }
 
 /*----------------------------------------------------------------------------*/
@@ -683,7 +542,7 @@ int knot_node_is_non_auth(const knot_node_t *node)
 		return KNOT_EINVAL;
 	}
 
-	return knot_node_flags_get_nonauth(node->flags);
+	return knot_node_flags_get(node, KNOT_NODE_FLAGS_NONAUTH);
 }
 
 /*----------------------------------------------------------------------------*/
@@ -694,8 +553,8 @@ void knot_node_set_auth(knot_node_t *node)
 		return;
 	}
 
-	knot_node_flags_clear_nonauth(&node->flags);
-	knot_node_flags_clear_deleg(&node->flags);
+	knot_node_flags_clear(node, KNOT_NODE_FLAGS_NONAUTH);
+	knot_node_flags_clear(node, KNOT_NODE_FLAGS_DELEG);
 }
 
 /*----------------------------------------------------------------------------*/
@@ -713,41 +572,72 @@ int knot_node_is_auth(const knot_node_t *node)
 
 int knot_node_is_empty(const knot_node_t *node)
 {
-	return knot_node_flags_get_empty(node->flags);
+	return knot_node_flags_get(node, KNOT_NODE_FLAGS_EMPTY);
 }
 
 /*----------------------------------------------------------------------------*/
 
 void knot_node_set_empty(knot_node_t *node)
 {
-	knot_node_flags_set_empty(&node->flags);
+	knot_node_flags_set(node, KNOT_NODE_FLAGS_EMPTY);
 }
 
 /*----------------------------------------------------------------------------*/
 
-static void knot_node_free_rrsets_from_tree(void *item, void *data)
+int knot_node_is_replaced_nsec(const knot_node_t *node)
 {
-	if (item == NULL) {
-		return;
-	}
-	
-	knot_rrset_t *rrset = (knot_rrset_t *)(item);
-	knot_rrset_deep_free(&rrset, 0, 1, *((int *)data));
+	return knot_node_flags_get(node, KNOT_NODE_FLAGS_REPLACED_NSEC);
 }
 
 /*----------------------------------------------------------------------------*/
 
-void knot_node_free_rrsets(knot_node_t *node, int free_rdata_dnames)
+void knot_node_set_replaced_nsec(knot_node_t *node)
+{
+	knot_node_flags_set(node, KNOT_NODE_FLAGS_REPLACED_NSEC);
+}
+
+/*----------------------------------------------------------------------------*/
+
+void knot_node_clear_replaced_nsec(knot_node_t *node)
+{
+	knot_node_flags_clear(node, KNOT_NODE_FLAGS_REPLACED_NSEC);
+}
+
+/*----------------------------------------------------------------------------*/
+
+void knot_node_set_apex(knot_node_t *node)
 {
 	if (node == NULL) {
 		return;
 	}
 
-	char *name = knot_dname_to_str(node->owner);
-	free(name);
+	knot_node_flags_set(node, KNOT_NODE_FLAGS_APEX);
+}
 
-	gen_tree_destroy(&node->rrset_tree, knot_node_free_rrsets_from_tree, 
-	                 (void *)&free_rdata_dnames);
+/*----------------------------------------------------------------------------*/
+
+int knot_node_is_apex(const knot_node_t *node)
+{
+	if (node == NULL) {
+		return KNOT_EINVAL;
+	}
+
+	return knot_node_flags_get(node, KNOT_NODE_FLAGS_APEX);
+}
+
+
+/*----------------------------------------------------------------------------*/
+
+void knot_node_free_rrsets(knot_node_t *node)
+{
+	if (node == NULL) {
+		return;
+	}
+
+	knot_rrset_t **rrs = node->rrset_tree;
+	for (uint16_t i = 0; i < node->rrset_count; ++i) {
+		knot_rrset_deep_free(&(rrs[i]), 1);
+	}
 }
 
 /*----------------------------------------------------------------------------*/
@@ -757,23 +647,17 @@ void knot_node_free(knot_node_t **node)
 	if (node == NULL || *node == NULL) {
 		return;
 	}
-	
+
 	dbg_node_detail("Freeing node: %p\n", *node);
 
 	if ((*node)->rrset_tree != NULL) {
 		dbg_node_detail("Freeing RRSets.\n");
-		gen_tree_destroy(&(*node)->rrset_tree, NULL, NULL);
+		free((*node)->rrset_tree);
+		(*node)->rrset_tree = NULL;
+		(*node)->rrset_count = 0;
 	}
 
-	// set owner's node pointer to NULL, but only if the 'node' does
-	// not point to the owner's node
-	if (node != &(*node)->owner->node
-	    && knot_dname_node(knot_node_owner(*node)) == *node) {
-		knot_dname_set_node((*node)->owner, NULL);
-	}
-
-	dbg_node_detail("Releasing owner.\n");
-	knot_dname_release((*node)->owner);
+	knot_dname_free(&(*node)->owner);
 
 	free(*node);
 	*node = NULL;
@@ -787,7 +671,7 @@ int knot_node_compare(knot_node_t *node1, knot_node_t *node2)
 {
 	assert(node1 != NULL && node2 != NULL);
 
-	return knot_dname_compare(node1->owner, node2->owner);
+	return knot_dname_cmp(node1->owner, node2->owner);
 }
 
 /*----------------------------------------------------------------------------*/
@@ -799,30 +683,26 @@ int knot_node_shallow_copy(const knot_node_t *from, knot_node_t **to)
 	}
 
 	// create new node
-	*to = knot_node_new(from->owner, NULL, from->flags);
+	*to = knot_node_new(NULL, NULL, from->flags);
 	if (*to == NULL) {
 		return KNOT_ENOMEM;
 	}
 
-	// set the parent by hand, so that the children count is not affected
-	(*to)->parent = from->parent;
-
-	/* Free old rrset_tree, as it will be replaced by shallow copy. */
-	gen_tree_destroy(&(*to)->rrset_tree, 0, 0);
-
-	// copy references	
-	// do not use the API function to set parent, so that children count 
+	// do not use the API function to set parent, so that children count
 	// is not changed
 	memcpy(*to, from, sizeof(knot_node_t));
+	(*to)->owner = knot_dname_copy(from->owner);
 
 	// copy RRSets
-	(*to)->rrset_tree = gen_tree_shallow_copy(from->rrset_tree);
-
+	size_t rrlen = sizeof(knot_rrset_t*) * from->rrset_count;
+	(*to)->rrset_tree = malloc(rrlen);
 	if ((*to)->rrset_tree == NULL) {
 		free(*to);
 		*to = NULL;
 		return KNOT_ENOMEM;
 	}
+	memcpy((*to)->rrset_tree, from->rrset_tree, rrlen);
 
 	return KNOT_EOK;
 }
+
