@@ -29,17 +29,14 @@
 
 #include <time.h>
 
-#include "zone/node.h"
-#include "dname.h"
-#include "nsec3.h"
-#include "zone/dname-table.h"
-#include "common/tree.h"
+#include "libknot/zone/node.h"
+#include "libknot/dname.h"
+#include "libknot/dnssec/nsec3.h"
 #include "common/ref.h"
-#include "hash/cuckoo-hash-table.h"
 
-#include "zone-tree.h"
+#include "libknot/zone/zone-tree.h"
 
-#include "zone/zone-contents.h"
+#include "libknot/zone/zone-contents.h"
 
 /*----------------------------------------------------------------------------*/
 /*!
@@ -58,9 +55,11 @@ typedef enum knot_zone_retvals knot_zone_retvals_t;
  * \brief Zone flags.
  */
 typedef enum knot_zone_flag_t {
-	KNOT_ZONE_SLAVE     = 0 << 0, /*! Slave zone */
-	KNOT_ZONE_MASTER    = 1 << 0, /*! Master zone. */
-	KNOT_ZONE_DISCARDED = 1 << 1  /*! Zone waiting to be discarded. */
+	KNOT_ZONE_SLAVE      = 0 << 0, /*! Slave zone */
+	KNOT_ZONE_MASTER     = 1 << 0, /*! Master zone. */
+	KNOT_ZONE_DISCARDED  = 1 << 1, /*! Zone waiting to be discarded. */
+	KNOT_ZONE_UPDATED    = 1 << 2, /*! Zone is updated in this cycle. */
+	KNOT_ZONE_OBSOLETE   = 1 << 3  /*! Zone is obsolete (forces retransfer). */
 } knot_zone_flag_t;
 
 /*----------------------------------------------------------------------------*/
@@ -105,12 +104,10 @@ knot_zone_t *knot_zone_new_empty(knot_dname_t *name);
  * \brief Creates new DNS zone.
  *
  * \param apex Node representing the zone apex.
- * \param node_count Number of authorative nodes in the zone.
  *
  * \return The initialized zone structure or NULL if an error occured.
  */
-knot_zone_t *knot_zone_new(knot_node_t *apex, uint node_count,
-                               int use_domain_table);
+knot_zone_t *knot_zone_new(knot_node_t *apex);
 
 knot_zone_contents_t *knot_zone_get_contents(
 	const knot_zone_t *zone);
@@ -151,11 +148,8 @@ void knot_zone_free(knot_zone_t **zone);
  * Also sets the given pointer to NULL.
  *
  * \param zone Zone to be freed.
- * \param free_rdata_dnames Set to <> 0 if you want to delete ALL domain names
- *                          present in RDATA. Set to 0 otherwise. (See
- *                          knot_rdata_deep_free().)
  */
-void knot_zone_deep_free(knot_zone_t **zone, int destroy_dname_table);
+void knot_zone_deep_free(knot_zone_t **zone);
 
 /*!
  * \brief Set destructor and initialize reference counter to 1.
@@ -193,7 +187,11 @@ void knot_zone_set_dtor(knot_zone_t *zone, int (*dtor)(struct knot_zone *));
  * \param zone Zone.
  */
 static inline unsigned knot_zone_flags(knot_zone_t *zone) {
-	return zone->flags;
+	if (zone) {
+		return zone->flags;
+	} else {
+		return 0;
+	}
 }
 
 /*!
